""" Module that contains Manager class. """
import os
import random
from typing import Callable, Iterable

import numpy as np
import torch


class Manager:
    """
    This class runs multiple experiments (one for each seed) sequentially.

    Parameters
    ----------
    seeds : iterable of ints
        Seeds to use in the experiments.
    """

    def __init__(
        self,
        seeds: Iterable[int],
        folder_name: str,
        deterministic: bool = False,
    ):
        self.seeds = seeds
        self.folder_name = folder_name
        self.deterministic = deterministic

        if str(self.folder_name) not in os.listdir():
            os.mkdir(self.folder_name)

    def run_experiments(self, experiment_func: Callable) -> None:
        """Run 'experiment_func' for each random seed.

        All the logic, including model and optimizer states saving, must be
        coded in the 'experiment_func' variable.

        'experiment_func' must receive 2 arguments: 'seed' and 'folder_name'.
        These arguments may be used by the user to create the necessary folders
        to save the experiment.

        Parameters
        ----------
        experiment_func : function
            Experiment function.
        """
        for seed in self.seeds:
            self.__set_seed(seed=seed)
            experiment_func(seed=seed, folder_name=self.folder_name)

    def __set_seed(self, seed: int) -> None:
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.use_deterministic_algorithms(self.deterministic)



""" PyTorch models fitting package. """

# relative subpackages import
from . import (
    callbacks,
    conventions,
    io,
    manager,
    testing,
    trainer,
    utils,
)

__all__ = [
    "io",
    "utils",
    "trainer",
    "testing",
    "manager",
    "callbacks",
    "conventions",
]

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



""" This class wraps functionality to train PyTorch models """
from ._trainer import Trainer
from ._utils import MetricsHandler, TrainerInternalState

__all__ = ["Trainer", "TrainerInternalState", "MetricsHandler"]



""" Internal conventions of torchfitter. """


class ParamsDict:
    """
    Naming conventions for torchfitter.trainer.Trainer internal parameters.

    Attributes
    ----------
    TRAIN_LOSS : str
        The current training loss.
    VAL_LOSS : str
        The current validation loss.
    BATCH_TRAIN_LOSS : str
        Current batch train loss.
    BATCH_VAL_LOSS : str
        Current batch validation loss.
    OPTIMIZER : str
        Algorithm used to optimize the model.
    EPOCH_TIME : str
        The time it took to compute the current epoch.
    EPOCH_NUMBER : str
        The corresponding number of the current epoch.
    TOTAL_EPOCHS : str
        The total number of epochs.
    TOTAL_TIME : str
        The total time it took to complete all epochs.
    STOP_TRAINING : str
        The total time it took to complete all epochs.
    DEVICE : str
        Device where the model and data are stored.
    MODEL : str
        The model to train.
    ACCELERATOR : str
        The accelerate.Accelerator object used to boost the training process.
    BATCH_HISTORY : str
        History of the batch metrics. It is a dictionary
        containing the passed and the following metrics:
        * ParamsDict.HISTORY_TRAIN_LOSS
        * ParamsDict.HISTORY_VAL_LOSS
        * ParamsDict.HISTORY_LR
    EPOCH_HISTORY : str
        History of the epoch (accumulated batch) metrics. It is a dictionary
        containing the passed and the following metrics:
        * ParamsDict.HISTORY_TRAIN_LOSS
        * ParamsDict.HISTORY_VAL_LOSS
        * ParamsDict.HISTORY_LR
    LOSS : str
        Loss criterion historical value. It is a dictionary containing the keys
        'train' and 'validation'.
    HISTORY_LR : str
        Learning rate for each epoch up to the current epoch.
    TRAIN_LOADER : str
        Train dataloader.
    TRAIN_BATCH : str
        Current training batch.
    TRAIN_BATCH_IDX : str
        Current training batch index.
    VAL_LOADER : str
        Validation dataloader.
    VAL_BATCH : str
        Current validation batch.
    VAL_BATCH_IDX : str
        Current validation batch index.
    """

    TRAIN_LOSS = "training_loss"
    VAL_LOSS = "validation_loss"
    BATCH_TRAIN_LOSS = "batch_training_loss"
    BATCH_VAL_LOSS = "batch_validation_loss"
    OPTIMIZER = "optimizer"
    EPOCH_TIME = "epoch_time"
    EPOCH_NUMBER = "epoch_number"
    TOTAL_EPOCHS = "total_epochs"
    TOTAL_TIME = "total_time"
    STOP_TRAINING = "stop_training"
    DEVICE = "device"
    MODEL = "model"
    ACCELERATOR = "accelerator"
    BATCH_HISTORY = "batch_history"
    EPOCH_HISTORY = "epoch_history"
    LOSS = "loss"
    HISTORY_LR = "learning_rate"
    TRAIN_LOADER = "train_loader"
    TRAIN_BATCH = "train_batch"
    TRAIN_BATCH_IDX = "train_batch_idx"
    VAL_LOADER = "validation_loader"
    VAL_BATCH = "val_batch"
    VAL_BATCH_IDX = "val_batch_idx"



"""
Pool of utilities to wrap data.
"""
from typing import Tuple, Union

import numpy as np
import torch
from torch.utils.data import Dataset

from torchfitter.utils.preprocessing import numpy_to_torch

__all__ = [
    "DataWrapper",
    "FastTensorDataLoader",
]


class DataWrapper(Dataset):
    """
    Class to wrap a dataset. Assumes X and y are torch.Tensors or numpy.arrays.
    The DataWrapper will access the elements by indexing the first axis.

    Parameters
    ----------
    X : torch.Tensor or numpy.array
        Features tensor.
    y : torch.Tensor or numpy.array
        Labels tensor.
    dtype_X : str, optional, default: 'float'
        Data type for features dataset.
    dtype_y : str, optional, default: 'int'
        Data type for labels dataset.
    """

    def __init__(
        self,
        X: torch.Tensor,
        y: torch.Tensor,
        dtype_X: str = "float",
        dtype_y: str = "int",
    ):
        X, y = self._check_inputs(X, y, dtype_X, dtype_y)

        self.features = X
        self.labels = y

    def __len__(self):
        return len(self.features)

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        return self.features[idx], self.labels[idx]

    def _check_inputs(
        self,
        X: Union[np.ndarray, torch.Tensor],
        y: Union[np.ndarray, torch.Tensor],
        dtype_X: float,
        dtype_y: float,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Check if the given inputs are numpy arrays and convert them if
        necessary.
        """
        if isinstance(X, np.ndarray):
            X = numpy_to_torch(X, dtype_X)
        else:
            X = getattr(X, dtype_X)()

        if isinstance(y, np.ndarray):
            y = numpy_to_torch(y, dtype_y)
        else:
            y = getattr(X, dtype_y)()

        return X, y


class FastTensorDataLoader:
    """DataLoader with faster loading.

    This class allows for a faster data loading. Although it won't always
    speed up the loading process, it can make the loading process 20 times
    faster.

    See `References` section to know the author.

    Parameters
    ----------
    *tensors : tuple of torch.Tensor
        Tensors to store.
    batch_size : int, optional, default: 32
        The batch size to load.
    shuffle : bool, optional, default: False
        Whether to shuffle the data (True) or not (False). If False, data will
        be processed in sequentially.

    References
    ----------
    .. [1] PyTorch discuss - Dataloader much slower than manual batching:
        https://discuss.pytorch.org/t/dataloader-much-slower-than-manual-
        batching/27014/6

    """

    def __init__(
        self,
        *tensors: torch.Tensor,
        batch_size: int = 32,
        shuffle: bool = False
    ):
        assert all(t.shape[0] == tensors[0].shape[0] for t in tensors)
        self.tensors = tensors

        self.dataset_len = self.tensors[0].shape[0]
        self.batch_size = batch_size
        self.shuffle = shuffle

        # Calculate # batches
        n_batches, remainder = divmod(self.dataset_len, self.batch_size)
        if remainder > 0:
            n_batches += 1

        self.n_batches = n_batches

    def __iter__(self):
        if self.shuffle:
            r = torch.randperm(self.dataset_len)
            self.tensors = [t[r] for t in self.tensors]
        self.i = 0

        return self

    def __next__(self) -> torch.Tensor:
        if self.i >= self.dataset_len:
            raise StopIteration

        start = self.i
        end = start + self.batch_size
        batch = tuple(t[start:end] for t in self.tensors)
        self.i += self.batch_size

        return batch

    def __len__(self) -> int:
        return self.n_batches



""" Module to handle input/output operations. """

import pickle

__all__ = ["save_pickle", "load_pickle"]


def save_pickle(obj, path, protocol=pickle.HIGHEST_PROTOCOL):
    """
    Save given object as pickle.

    Parameters
    ----------
    obj : object
        Object to save.
    path : str or Path
        Path where to save the pickle.
    protocol : int, optional, default: pickle.HIGHEST_PROTOCOL
        Used pickle protocol.

    """
    with open(path, "wb") as handle:
        pickle.dump(obj, handle, protocol=protocol)


def load_pickle(path):
    """
    Load saved pickle.

    Parameters
    ----------
    path : str or Path
        Path where to save the pickle.

    Returns
    -------
    obj : object
        Loaded object.

    """
    with open(path, "rb") as handle:
        obj = pickle.load(handle)

    return obj



import logging
import statistics
import time
from typing import List, Tuple, Union

import torch
import torchmetrics
from accelerate import Accelerator
from numpy import ndarray
from torch.utils.data.dataloader import DataLoader

from torchfitter.callbacks.base import Callback, CallbackHandler
from torchfitter.conventions import ParamsDict
from torchfitter.trainer._utils import MetricsHandler, TrainerInternalState


class Trainer:
    """Class that eases the training of a PyTorch model.

    This class leverages the power of 'accelerate' to handle the device
    management and the model optimization.

    The trainer tracks its state using a
    'torchfitter.trainer.TrainerInternalState' object. You can also pass a list
    of callbacks and/or metrics to the trainer. The callbacks will be runned
    at different points depending on the methods that were filled. The metrics
    will be runned in the train and validation steps.

    Parameters
    ----------
    model : torch.Module
        The model to train.
    criterion : torch.Module
        Loss function criterion used to optimize the model.
    optimizer : torch.optim
        Optimizer to perform the parameters update.
    mixed_precision : bool, optional, default: False
        Whether to use mixed precision training or not. If True, the forward
        pass will be computed under the context of `torch.cuda.amp.autocast`
        and the backpropagation and gradient descent steps will be computed
        using `torch.cuda.amp.GradScaler`.
    callbacks : list of torchfitter.callback.Callback
        Callbacks to use during the training process.
    metrics : list of torchmetrics.Metric, optional, default: None
        List of metrics to compute in the fitting process. The metrics will be
        registered in the internal state using the name of the class. For
        example, passing `[MeanSquaredError()]` will be registered as
        `MeanSquaredError`.
    accelerator : accelerate.Accelerator
        Accelerator object from 'accelerate'. If no object is passed, the
        trainer will create an instance with the default parameters.
    accumulate_iter : int, optional, default: 1
        Accumulate gradients every 'accumulate_iter' iterations. The default
        value does not accumulate the gradients. If an instance of Accelerator
        is passed to the trainer, this parameter will be ignored.
    gradient_clipping : {None, 'norm', 'value'}
        Norm gradient clipping of value gradient clipping. If None, gradient
        clipping won't be applied.
    gradient_clipping_kwargs : dict, optional, default: None
        Dictionary containing keyword arguments for gradient clipping
        algorithm. Example: {max_norm=1, norm_type=2}. See
        https://huggingface.co/docs/accelerate/accelerator.html for more
        information.
    log_level : int, optional, default: logging.INFO
        Logging level.

    Attributes
    ----------
    callback_handler : torchfitter.callback.CallbackHandler
        Handles the passed callbacks.
    internal_state : torchfitter.trainer.TrainerInternalState
        Trainer internal parameters state.
    metrics_handler : torchfitter.trainer.MetricsHandler
        Handles the passed metrics.
    gradient_clipping_algo_ : callable
        Gradient clipping algorithm or None.

    """

    def __init__(
        self,
        model: torch.nn.Module,
        criterion: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        mixed_precision: bool = False,
        callbacks: List[Callback] = None,
        metrics: List[torchmetrics.Metric] = None,
        accelerator: Accelerator = None,
        accumulate_iter: int = 1,
        gradient_clipping: str = None,
        gradient_clipping_kwargs: dict = None,
        log_level: int = logging.INFO,
    ):
        self.criterion = criterion
        self.callbacks_list = callbacks
        self.metrics_list = metrics
        self.accumulate_iter = accumulate_iter
        self.gradient_clipping = gradient_clipping
        self.gradient_clipping_kwargs = gradient_clipping_kwargs
        self.log_level = log_level

        if accelerator is None:
            self.accelerator = Accelerator(
                fp16=mixed_precision,
                gradient_accumulation_steps=accumulate_iter,
            )

        # wrap withing accelerate environment
        self.optimizer = self.accelerator.prepare_optimizer(optimizer)
        self.model = self.accelerator.prepare_model(model)

        # ----- attributes -----
        self.internal_state = TrainerInternalState(
            model=self.model, accelerator=self.accelerator, optimizer=optimizer
        )
        self.callback_handler = CallbackHandler(
            callbacks_list=self.callbacks_list
        )
        self.callback_handler.set_log_level(self.log_level)

        self.metrics_handler = MetricsHandler(
            metrics_list=self.metrics_list,
            criterion=criterion,
            device=self.internal_state.get_single_param(ParamsDict.DEVICE),
        )
        self.gradient_clipping_algo_ = self._prepare_gradient_clipping()

        if self.metrics_handler.metric_names is not None:
            names = self.metrics_handler.metric_names
            self.internal_state.add_metrics(*names)

    def fit(
        self,
        train_loader: DataLoader,
        val_loader: DataLoader,
        epochs: int,
    ) -> dict:
        """Fit the model.

        Fit the model using the given loaders for the given number of epochs.
        By default, the trainer does not display any information about the
        fitting process, but you can use any of the callbacks designed for that
        purpose or create your own callbacks.

        Parameters
        ----------
        train_loader : torch.DataLoader
            DataLoader to iterate over the train dataset.
        val_loader : torch.DataLoader
            DataLoader to iterate over the validation dataset.
        epochs : int
            Number of training epochs.

        Returns
        -------
        history : dict
            Dictionay with epoch and batch metrics. The metrics contained in
            the dictionary will be the passed metrics + criterion results.

        """
        initial_epoch = self.internal_state.get_single_param(
            key=ParamsDict.EPOCH_NUMBER
        )

        # wrap loaders within accelerate environment
        train_loader = self.accelerator.prepare_data_loader(train_loader)
        val_loader = self.accelerator.prepare_data_loader(val_loader)

        # update internal state with loaders
        self.internal_state.update_params(
            **{
                ParamsDict.TRAIN_LOADER: train_loader,
                ParamsDict.VAL_LOADER: val_loader,
                ParamsDict.TOTAL_EPOCHS: epochs,
            }
        )

        # track total training time
        total_start_time = time.perf_counter()
        self.callback_handler.on_fit_start(self.state_dict())

        # ---- fitting process ----
        epoch = initial_epoch
        stop = False
        while epoch <= epochs and not stop:
            self.callback_handler.on_epoch_start(self.state_dict())

            # track epoch time
            epoch_start_time = time.perf_counter()

            # ------- train step -------
            self.callback_handler.on_train_step_start(self.state_dict())
            tr_loss = self.train_step(train_loader)  # actual step
            self.callback_handler.on_train_step_end(self.state_dict())

            # ------- validation step -------
            self.callback_handler.on_validation_step_start(self.state_dict())
            val_loss = self.validation_step(val_loader)
            self.callback_handler.on_validation_step_end(self.state_dict())

            # -------- update internal state to track training --------
            self.internal_state.update_lr_history(
                value=self.optimizer.param_groups[0]["lr"], is_batch=False
            )

            # synchronize before measuring time
            self.accelerator.wait_for_everyone()
            epoch_time = time.perf_counter() - epoch_start_time
            self.internal_state.update_params(
                **{
                    ParamsDict.VAL_LOSS: val_loss,
                    ParamsDict.TRAIN_LOSS: tr_loss,
                    ParamsDict.EPOCH_TIME: epoch_time,
                }
            )

            self.callback_handler.on_epoch_end(self.state_dict())

            epoch += 1
            stop = self.internal_state.get_single_param(
                key=ParamsDict.STOP_TRAINING
            )
            self.internal_state.update_params(
                **{ParamsDict.EPOCH_NUMBER: epoch}
            )

        total_time = time.perf_counter() - total_start_time

        self.internal_state.update_params(
            **{ParamsDict.TOTAL_TIME: total_time}
        )
        self.callback_handler.on_fit_end(self.state_dict())

        history = self.get_history()
        return history

    @torch.no_grad()
    def predict(
        self,
        X: Union[DataLoader, torch.Tensor, ndarray],
        as_array=False,
        dtype: str = "float",
    ) -> Union[torch.Tensor, ndarray]:
        """
        Predict function.

        Parameters
        ----------
        X : torch.Tensor or numpy.ndarray
            Data to use to make inference.
        as_array : bool, optional, default: False
            Whether to output the predictions as a numpy.narray or not.
        dtype : str, optional, default: "float"
            Data type to cast input tensor to.

        Returns
        -------
        predictions : torch.Tensor or numpy.ndarray
            Predicted values.
        """
        if isinstance(X, DataLoader):
            _tensor = self.__predict_loader(X)
            predictions = getattr(_tensor, dtype)()

        elif isinstance(X, ndarray):
            _numpy = torch.from_numpy(X)
            X = getattr(_numpy, dtype)()
            predictions = self.__predict_tensor(X)

        elif isinstance(X, torch.Tensor):
            _tensor = self.__predict_tensor(X)
            predictions = getattr(_tensor, dtype)()

        if as_array:
            return predictions.cpu().numpy()
        else:
            return predictions

    def __predict_tensor(self, tensor: torch.Tensor) -> torch.Tensor:
        """Make prediction for a given torch tensor.

        The passed tensor will be moved to the device the accelerator chose at
        the beginning of the training process.

        Parameters
        ----------
        tensor : torch.Tensor
            Tensor to use to make inference.

        Returns
        -------
        torch.Tensor
            Predicted values.
        """
        device = self.accelerator.device
        tensor = tensor.to(device)
        return self.model(tensor)

    def __predict_loader(self, loader: DataLoader) -> torch.Tensor:
        """Make inference prediction for a given torch.DataLoader.

        Useful when the tensor of features does not fit into memory.
        """
        _predictions = []
        loader = self.accelerator.prepare_data_loader(loader)
        for idx, (feat, lab) in enumerate(loader):
            _pred = self.model(feat)
            _predictions.append(_pred)

        predictions = torch.cat(_predictions)
        return predictions

    def _prepare_gradient_clipping(self) -> callable:
        """
        Identify the gradient clipping algorithm to use.

        Returns
        -------
        algo : callable
            Callable function that wraps the gradient clipping funcionality.
        """
        if self.gradient_clipping == "value":
            algo = self.accelerator.clip_grad_value_

        elif self.gradient_clipping == "norm":
            algo = self.accelerator.clip_grad_norm_

        elif self.gradient_clipping is None:
            algo = None

        else:
            raise ValueError(
                "Not supported gradient "
                f"gradient clipping algorithm: '{self.gradient_clipping}'"
            )
        return algo

    def set_scaler(
        self, scaler: torch.cuda.amp.grad_scaler.GradScaler
    ) -> None:
        """Set the gradient scaler used in mixed precision.

        Since the trainer relies on accelerate.Accelator class for the fitting
        process the scaler is created inside the aforementioned. With this
        function you can set the scaler to be an instance with the desired
        argument values.

        Parameters
        ----------
        scaler : torch.cuda.amp.grad_scaler.GradScaler
            The gradient scaler you want to use.
        """
        self.accelerator.scaler = scaler

    def reset_parameters(self, reset_model: bool = False) -> None:
        """
        Reset the internal dictionary that keeps track of the parameters state.

        Parameters
        ----------
        reset_model : bool, optional, default: False
            True to reset the model state.
        """
        restart_dict = self.internal_state.reset_parameters(
            reset_model=reset_model
        )
        self.params_dict = restart_dict

    def train_step(
        self, loader: torch.utils.data.dataloader.DataLoader
    ) -> float:
        """Perform a train step using the given dataloader.

        A train step consists of running and optimizing the model for each
        batch in the given train dataloader.

        Parameters
        ----------
        loader: torch.utils.data.dataloader.DataLoader
            Dataloader for train set.

        Returns
        -------
        epoch_loss : float
            Mean loss of the batch.
        """
        self.model.train()

        losses = []  # loss as mean of batch losses
        for batch_idx, batch in enumerate(loader):
            self.callback_handler.on_train_batch_start(self.state_dict())
            loss = self.batch_train_step(batch_index=batch_idx, batch=batch)
            self.callback_handler.on_train_batch_end(self.state_dict())

            losses.append(loss.item())

        # compute accumulated metrics (metric.compute())
        metrics_accumulated = (
            self.metrics_handler.accumulated_batch_computation()
        )

        if metrics_accumulated is not None:
            self.internal_state.update_metrics(
                is_train=True, is_batch=False, **metrics_accumulated
            )

        epoch_loss = statistics.mean(losses)
        self.internal_state.update_loss_history(
            value=epoch_loss, is_train=True, is_batch=False
        )

        # reset metrics
        self.metrics_handler.reset_metrics()
        return epoch_loss

    def batch_train_step(
        self,
        batch_index,
        batch: Tuple[torch.Tensor, torch.Tensor],
    ) -> torch.Tensor:
        """
        Define the computations to perform on each batch for the training loop.

        Parameters
        ----------
        batch_index : int
            Batch index.
        batch : tuple
            Batch to process.

        Returns
        -------
        loss : torch.Tensor
            Train loss graph.
        """
        with self.accelerator.accumulate(self.model):
            # assume last tensor in batch are labels
            batch_len = len(batch)
            features, labels = batch[: batch_len - 1], batch[-1]

            # forward propagation
            out = self.model(*features)
            loss = self.loss_step(out, labels, is_validation=False)

            # backpropagation
            self.accelerator.backward(loss)

            # gradient clipping
            if self.gradient_clipping_algo_ is not None:
                self.gradient_clipping_algo_(
                    self.model.parameters(), **self.gradient_clipping_kwargs
                )

            self.optimizer.step()
            self.optimizer.zero_grad()

        # compute metrics, needed for accumulated computation
        metrics_single = self.metrics_handler.single_batch_computation(
            predictions=out, target=labels
        )

        # ----------- internal state updated ------------
        if metrics_single is not None:
            self.internal_state.update_metrics(
                is_train=True, is_batch=True, **metrics_single
            )

        # update history
        self.internal_state.update_loss_history(
            value=loss.item(), is_train=True, is_batch=True
        )

        self.internal_state.update_lr_history(
            value=self.optimizer.param_groups[0]["lr"], is_batch=True
        )

        # update internal state
        self.internal_state.update_params(
            **{
                ParamsDict.TRAIN_BATCH: batch,
                ParamsDict.TRAIN_BATCH_IDX: batch_index,
            }
        )

        return loss

    @torch.no_grad()
    def validation_step(
        self, loader: torch.utils.data.dataloader.DataLoader
    ) -> float:
        """Perform a validation step using the given dataloader.

        A validation step consists of running the model for each batch in the
        given validation dataloader.

        This method runs under the context of "torch.no_grad", which means
        gradients won't be tracked.

        Parameters
        ----------
        loader: torch.utils.data.dataloader.DataLoader
            Dataloader for validation set.

        Returns
        -------
        epoch_loss : float
            Mean loss of the batch.
        """
        self.model.eval()

        losses = []  # loss as mean of batch losses
        for batch_idx, batch in enumerate(loader):
            self.callback_handler.on_validation_batch_start(self.state_dict())
            loss = self.batch_validation_step(
                batch_index=batch_idx, batch=batch
            )
            self.callback_handler.on_validation_batch_end(self.state_dict())
            losses.append(loss.item())

        # compute accumulated metrics
        metrics_accumulated = (
            self.metrics_handler.accumulated_batch_computation()
        )

        if metrics_accumulated is not None:
            self.internal_state.update_metrics(
                is_train=False, **metrics_accumulated
            )

        epoch_loss = statistics.mean(losses)
        self.internal_state.update_loss_history(
            value=epoch_loss, is_train=False, is_batch=False
        )

        # reset metrics
        self.metrics_handler.reset_metrics()

        return epoch_loss

    def batch_validation_step(
        self, batch_index, batch: Tuple[torch.Tensor, torch.Tensor]
    ) -> torch.Tensor:
        """
        Define the computations to perform on each batch for the validation
        loop.

        Parameters
        ----------
        batch_index : int
            Batch index.
        batch : tuple
            Batch to process.

        Returns
        -------
        loss : torch.Tensor
            Validation loss graph.
        """
        # assume last tensor in batch are labels
        batch_len = len(batch)
        features, labels = batch[: batch_len - 1], batch[-1]

        out = self.model(*features)
        loss = self.loss_step(out, labels, is_validation=True)

        # compute metrics, needed for accumulated computation
        metrics_single = self.metrics_handler.single_batch_computation(
            predictions=out, target=labels
        )

        # ----------- internal state updated ------------
        # update metrics dict
        if metrics_single is not None:
            self.internal_state.update_metrics(
                is_train=False, is_batch=True, **metrics_single
            )

        # update internal state
        self.internal_state.update_loss_history(
            value=loss.item(), is_train=False, is_batch=True
        )

        self.internal_state.update_lr_history(
            value=self.optimizer.param_groups[0]["lr"], is_batch=True
        )

        self.internal_state.update_params(
            **{
                ParamsDict.VAL_BATCH: batch,
                ParamsDict.VAL_BATCH_IDX: batch_index,
            }
        )

        return loss

    def loss_step(
        self, real: torch.Tensor, target: torch.Tensor, is_validation: bool
    ) -> torch.Tensor:
        """Compute loss graph.

        If you override this method the passed regularizer algorithms won't be
        applied unless they are specifically added.

        Parameters
        ----------
        real : torch.Tensor
            Obtained tensor after performing a forward pass.
        target : torch.Tensor
            Target tensor to compute loss.

        Returns
        -------
        loss : torch.Tensor
            Loss graph contained in a (1 x 1) torch.Tensor.
        """
        self.callback_handler.on_loss_step_begin(self.state_dict())
        with self.accelerator.autocast():
            loss = self.criterion(real, target)

        # select key to update
        if is_validation:
            key = ParamsDict.BATCH_VAL_LOSS
        else:
            key = ParamsDict.BATCH_TRAIN_LOSS

        # store loss graph
        self.internal_state.update_params(**{key: loss})

        # callback and retrieval in case loss was modified
        self.callback_handler.on_loss_step_end(self.state_dict())
        loss = self.internal_state.get_single_param(key)

        return loss

    def save_model(self, path):
        """
        Convenient method to save the model ensuring it is unwrapped and
        all processes are done.

        Parameters
        ----------
        path : path-like
            Path to save the model.
        """
        self.accelerator.wait_for_everyone()
        unwrapped_model = self.accelerator.unwrap_model(self.model)
        self.accelerator.save(unwrapped_model.state_dict(), path)

    def load_model(self, path):
        """
        Convenient method to load the model ensuring it is unwrapped.

        Parameters
        ----------
        path : path-like
            Path to load the model from.
        """
        unwrapped_model = self.accelerator.unwrap_model(self.model)
        unwrapped_model.load_state_dict(torch.load(path))
        self.model = unwrapped_model

    def state_dict(self) -> dict:
        """Return current state dict.

        The state dict will change as the training progresses.

        Returns
        -------
        state : dict
            A dictionary containing the current state of the trainer.
        """
        state = self.internal_state.get_state_dict()
        return state

    def get_history(self) -> dict:
        """Return the training history.

        The history will be created up to the last epoch.

        Returns
        -------
        history : dict
            Dictionary containing the history up to the last epoch.
        """
        history = {
            ParamsDict.EPOCH_HISTORY: self.internal_state.get_single_param(
                key=ParamsDict.EPOCH_HISTORY
            ),
            ParamsDict.BATCH_HISTORY: self.internal_state.get_single_param(
                key=ParamsDict.BATCH_HISTORY
            ),
        }
        return history



""" Abstractions to interact with models. """

from . import base
from ._callbacks import (
    EarlyStopping,
    GPUStats,
    LearningRateScheduler,
    LoggerCallback,
    RichProgressBar,
    StochasticWeightAveraging,
    L1Regularization,
    L2Regularization,
    ElasticNetRegularization,
)

__all__ = [
    "base",
    "GPUStats",
    "EarlyStopping",
    "LoggerCallback",
    "RichProgressBar",
    "LearningRateScheduler",
    "StochasticWeightAveraging",
    "L1Regularization",
    "L2Regularization",
    "ElasticNetRegularization",
]



""" Util functions for testing purposes. """
from typing import Iterable

import torch
import torch.nn as nn


def change_model_params(
    model: torch.nn.Module, weights: torch.Tensor, biases: torch.Tensor
) -> None:
    """
    Change the model parameters. There is not need to return the model since
    the update will happen on the original object.

    Parameters
    ----------
    model : nn.Module
        Model to update.
    weights : torch.Tensor
        New weights.
    biases : torch.Tensor
        New biases.
    """
    with torch.no_grad():
        model.weight = nn.Parameter(weights)
        model.bias = nn.Parameter(biases)


def compute_forward_gradient(module: torch.nn.Module, *tensors) -> dict:
    """Computes forward gradient.

    This function helps to test the gradient computation of a nn.Module class.

    It computes the gradients for the given variables. The result of a forward
    must be a torch.Tensor with the shape (1,1), otherwise is not possible to
    compute the gradients.

    Parameters
    ----------
    module : torch.nn.Module
        The module function with a defined forward.
    *tensors : tuple
        The input tensors to compute the gradients of.

    Returns
    -------
    gradients : dict
        A dictionary whose keys are a list of sorted integers and whose values
        are the gradients for each passed tensor (in the same order they were
        passed).

    Note
    ----
    This function returns the gradients for the leaf variables, not the
    intermediate gradients.

    """
    # minimal check
    for t in tensors:
        if not t.requires_grad:
            raise ValueError("Tensors must have 'requires_grad' activated.")

    # create computational graph
    forward = module(*tensors)

    # compute gradients
    if forward.shape != torch.Size([1]) and forward.shape != torch.Size([]):
        raise ValueError(
            "The passed tensors produced a vector result instead of a scalar "
            "result. It is not possible to compute the backward pass if the "
            "forward result is not a (1,1) torch.Tensor."
        )
    else:
        forward.backward()

    # store gradient for each passed tensor
    gradients = {}
    for i, t in enumerate(tensors):
        gradients[i] = t.grad

    return gradients


def check_monotonically_decreasing(
    iterable: Iterable[float], strict: bool = False
) -> bool:
    """Check if the given iterable is monotonically decreasing.

    The function allows to check strictly (all i + 1 are greater than i) or
    non-strictly (all i + 1 are greater or equal than i).

    Parameters
    ----------
    iterable : array-like
        Iterable object to check.
    strict : bool, optional, default: False
        Whether to strictly check monotonically decreasing (True) or not
        (False).

    Returns
    -------
    bool
        True if 'iterable' is monotonically decreasing.

    References
    ----------
    .. [1] Python - How to check list monotonicity
       https://stackoverflow.com/questions/4983258/python-how-to-check-list-
       monotonicity
    """
    _zip = zip(iterable, iterable[1:])
    if strict:
        return all(next_ > current for next_, current in _zip)
    else:
        return all(next_ >= current for next_, current in _zip)



""" Callbacks for the manager class """
import subprocess
from typing import List

import torch
from rich.progress import BarColumn, Progress, TimeRemainingColumn
from torch.optim.swa_utils import AveragedModel

from torchfitter.conventions import ParamsDict
from torchfitter.callbacks.base import Callback


class EarlyStopping(Callback):
    """Callback to handle early stopping.

    `EarlyStopping` will be performed on the validation loss (as it should be).
    The best observed model will be loaded if `load_best` is True.

    Paramaters
    ----------
    patience : int, optional, default: 50
        Number of epochs to wait after min has been reached. After 'patience'
        number of epochs without improvemente, the training stops.
    load_best : bool, optional, deafult: True
        Whether to load the best observed parameters (True) or not (False).
    path : str or Path, optional, default: 'checkpoint.pt'
        Path to store the best observed parameters.
    """

    def __init__(self, patience=50, load_best=True, path="checkpoint.pt"):
        super(EarlyStopping, self).__init__()
        self.path = path
        self.patience = patience
        self.load_best = load_best

        self.log_name = "EarlyStopping"

    def __repr__(self) -> str:
        return (
            f"EarlyStopping(patience={self.patience}, "
            f"load_best={self.load_best})"
        )

    def on_fit_start(self, params_dict: dict) -> None:
        self.wait = 0
        self.stopped_epoch = 0
        self.best = float("inf")

    def on_epoch_end(self, params_dict: dict) -> None:
        current_loss = params_dict[ParamsDict.VAL_LOSS]
        epoch_number = params_dict[ParamsDict.EPOCH_NUMBER]
        model = params_dict[ParamsDict.MODEL]

        if current_loss < self.best:
            self.best = current_loss
            self.wait = 0
            # save weights
            best_params = model.state_dict().copy()
            torch.save(best_params, self.path)
        else:
            self.wait += 1

            if self.wait >= self.patience:
                self.stopped_epoch = epoch_number
                # send signal to stop training
                params_dict[ParamsDict.STOP_TRAINING] = True
                # load best weights
                if self.load_best:
                    best_params = torch.load(self.path)
                    model.load_state_dict(best_params)
                    self.logger.info("Best observed parameters loaded.")

    def on_fit_end(self, params_dict: dict) -> None:
        if self.stopped_epoch > 0:
            self.logger.info(
                f"Early stopping applied at epoch: {self.stopped_epoch}"
            )


class LoggerCallback(Callback):
    """
    `LoggerCallback` is used to log some of the parameters dict in order to
    monitor the fitting process. The logged parameters are:
        - Current epoch.
        - Number of epochs.
        - Train loss.
        - Validation loss.
        - Time / epoch.

    It also outputs the total training time once the fitting process ends.

    Parameters
    ----------
    update_step : int, optional, default: 50
        Logs will be performed every 'update_step'.
    precision : int, optional, default: 2
        Number of decimals in the numbers.
    """

    def __init__(self, update_step: int, precision: int = 2):
        super(LoggerCallback, self).__init__()
        self.update_step = update_step
        self.prec = precision

        self.log_name = "LoggerCallback"

    def __repr__(self) -> str:
        return (
            f"LoggerCallback(update_step={self.update_step}, "
            f"precision={self.prec})"
        )

    def on_fit_start(self, params_dict: dict) -> None:
        dev = params_dict[ParamsDict.DEVICE]
        self.logger.info(f"Starting training process on {dev}")

    def on_epoch_end(self, params_dict: dict) -> None:
        epoch_number = params_dict[ParamsDict.EPOCH_NUMBER]
        total_epochs = params_dict[ParamsDict.TOTAL_EPOCHS]
        val_loss = params_dict[ParamsDict.VAL_LOSS]
        train_loss = params_dict[ParamsDict.TRAIN_LOSS]
        epoch_time = params_dict[ParamsDict.EPOCH_TIME]

        prec = self.prec
        msg = (
            f"Epoch {epoch_number}/{total_epochs} | Train loss: "
            f"{train_loss:.{prec}e} | Validation loss {val_loss:.{prec}e} | "
            f"Time/epoch: {epoch_time:.{prec}e} s"
        )

        if epoch_number % self.update_step == 0 or epoch_number == 1:
            self.logger.info(msg)

    def on_fit_end(self, params_dict: dict) -> None:
        total_time = params_dict[ParamsDict.TOTAL_TIME]
        # final message
        self.logger.info(
            f"""End of training. Total time: {total_time:0.5f} seconds"""
        )


class LearningRateScheduler(Callback):
    """Callback to schedule learning rate.

    `LearningRateScheduler` provides an easy abstraction to schedule the
    learning rate of the optimizer by calling `scheduler.step()` after the
    training step has been performed; i.e., `on_train_step_end`.

    If you were to use a learning rate scheduler in addition to stochastic
    averaging, you must pass both to an instance of `StochasticWeightAveraging`
    callback instance of creating an instance of `LearningRateScheduler`.

    Parameters
    ----------
    scheduler : torch.optim.lr_scheduler._LRScheduler
        Scheduler use to perform the lr reduction.
    metric : str, optional, default : torchfitter.conventions.ParamsDict.LOSS
        Metric to track in order to reduce the learning rate. If you want to
        use one of the passed metrics you must pass the class name. See
        examples section.
    on_train : bool, optional, default: True
        Whether to watch the train version of the metric (True) or the
        validation version of the metric (False)

    Examples
    --------
    Assume we already have the `kwargs` for the trainer and the necessary
    imports:

    >>> import torchmetrics
    >>> from torch.optim.lr_scheduler import ReduceLROnPlateau
    >>> from torchfitter.callbacks import LearningRateScheduler
    >>> sch = ReduceLROnPlateau(optimizer, factor=0.1, patience=50)

    The default metric is the loss. You can choose the validation or the
    training loss or you can pass another metric by doing:

    >>> lr_sch = LearningRateScheduler(
    ...     scheduler=sch, metric='MeanSquaredError', on_train=False)
    >>> metrics = [torchmetrics.MeanSquaredError]
    >>> trainer = Trainer(callbacks=[lr_sch], metrics=metrics, **kwargs)

    References
    ----------
    ..[1] PyTorch - How to adjust learning rate
       https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate

    """

    def __init__(
        self,
        scheduler: torch.optim.lr_scheduler._LRScheduler,
        metric: str = ParamsDict.LOSS,
        on_train: bool = True,
    ):
        super(LearningRateScheduler, self).__init__()

        self.scheduler = scheduler
        self.metric = metric
        self.on_train = on_train

    def __repr__(self) -> str:
        sch = type(self.scheduler).__name__
        return (
            f"LearningRateScheduler(scheduler={sch}, metric={self.metric}), "
            f"on_train={self.on_train}"
        )

    def on_fit_start(self, params_dict: dict) -> None:
        accelerator = params_dict[ParamsDict.ACCELERATOR]
        self.scheduler = accelerator.prepare(self.scheduler)

    def on_train_step_end(self, params_dict: dict) -> None:
        if self.metric is not None:
            key = "train" if self.on_train else "validation"
            epoch_hist = params_dict[ParamsDict.EPOCH_HISTORY]
            epoch_number = params_dict[ParamsDict.EPOCH_NUMBER]

            # avoid failing in first epoch when on_train=False
            if epoch_number > 1:
                metric = epoch_hist[self.metric][key][-1]
                self.scheduler.step(metric)
        else:
            self.scheduler.step()


class GPUStats(Callback):
    """GPU stats logger.

    The list of available queries can be found on NVIDIA smi queries. See
    `Notes` section for more information.

    Parameters
    ----------
    queries : list of str
        List of queries to log
    format : str
        Queries format.
    update_step : int, optional, default: 50
        Logs will be performed every 'update_step'.

    Notes
    -----
    To check the list of available queries, see
    https://nvidia.custhelp.com/app/answers/detail/a_id/3751/~/useful-nvidia-smi-queries

    """

    def __init__(
        self,
        format: str = "csv, nounits, noheader",
        queries: List[str] = [
            "name",
            "temperature.gpu",
            "utilization.gpu",
            "utilization.memory",
            "memory.used",
        ],
        update_step=50,
    ):
        super(GPUStats, self).__init__()

        self.queries = queries
        self.format = format
        self.update_step = update_step

        self.log_name = "GPU Stats"

    def __repr__(self) -> str:
        return (
            f"GPUStats(format={self.format}, queries={self.queries}, "
            f"queries={self.queries})"
        )

    def on_epoch_end(self, params_dict: dict) -> None:
        epoch_number = params_dict[ParamsDict.EPOCH_NUMBER]

        if epoch_number == 1 or epoch_number % self.update_step == 0:
            stdout = self._get_queries(
                queries=self.queries, format=self.format
            )
            msg = " | ".join(map(str, stdout))  # unpack and format
            self.logger.info(msg)

    def _get_queries(self, queries, format):
        stdout = []

        for query in queries:
            out = (
                subprocess.run(
                    f"nvidia-smi --query-gpu={query} --format={format}",
                    stdout=subprocess.PIPE,
                    encoding="utf-8",
                )
                .stdout.replace("\r", " ")
                .replace("\n", " ")
            )

            stdout.append(out)

        return stdout


class RichProgressBar(Callback):
    """
    This callback displays a progress bar to report the state of the training
    process: on each epoch, a new bar will be created and stacked below the
    previous bars.

    Metrics are logged using the library logger.

    Parameters
    ----------
    display_step : int
        Number of epochs to wait to display the progress bar.
    precision : int, optional, default: 2
        Number of decimals to use in the log.
    log_lr : bool, optional, default: False
        Whether to log the learning rate (True) or not (False).
    """

    def __init__(
        self, display_step: int = 1, log_lr: bool = False, precision: int = 2
    ):
        super(RichProgressBar, self).__init__()

        self.display_step = display_step
        self.prec = precision
        self.log_lr = log_lr

        self.log_name = "Rich Bar"

    def __repr__(self) -> str:
        return (
            f"RichProgressBar(display_step={self.display_step}, "
            f"log_lr={self.log_lr}, precision={self.precision})"
        )

    def on_fit_start(self, params_dict: dict) -> None:
        dev = params_dict[ParamsDict.DEVICE]
        self.logger.info(f"Starting training process on {dev}\n")

    def on_train_batch_end(self, params_dict: dict) -> None:
        epoch = params_dict[ParamsDict.EPOCH_NUMBER]
        accelerator = params_dict[ParamsDict.ACCELERATOR]

        if epoch % self.display_step == 0 or epoch == 1:
            accelerator.wait_for_everyone()
            self.progress_bar.advance(self.epoch_task, 1)

    def on_validation_batch_end(self, params_dict: dict) -> None:
        epoch = params_dict[ParamsDict.EPOCH_NUMBER]
        accelerator = params_dict[ParamsDict.ACCELERATOR]

        if epoch % self.display_step == 0 or epoch == 1:
            accelerator.wait_for_everyone()
            self.progress_bar.advance(self.epoch_task, 1)

    def on_epoch_start(self, params_dict: dict) -> None:
        # gather necessary objects
        train_loader = params_dict[ParamsDict.TRAIN_LOADER]
        val_loader = params_dict[ParamsDict.VAL_LOADER]
        epoch = params_dict[ParamsDict.EPOCH_NUMBER]
        total_epochs = params_dict[ParamsDict.TOTAL_EPOCHS]

        # compute number of batches
        n_elements = len(train_loader) + len(val_loader)

        if epoch % self.display_step == 0 or epoch == 1:
            self.progress_bar = Progress(
                "[progress.description]{task.description}",
                "•",
                BarColumn(),
                "•",
                "[progress.percentage]{task.percentage:>3.0f}%",
                "•",
                TimeRemainingColumn(),
            )

            self.epoch_task = self.progress_bar.add_task(
                description=f"Epoch {epoch}/{total_epochs}",
                total=n_elements,
            )
            self.progress_bar.start()

    def on_epoch_end(self, params_dict: dict) -> None:
        epoch = params_dict[ParamsDict.EPOCH_NUMBER]

        if epoch % self.display_step == 0 or epoch == 1:
            # update metrics
            text = self.render_text(params_dict[ParamsDict.EPOCH_HISTORY])
            self.logger.info(text)  # DISC: use included Rich logger?
            self.progress_bar.stop()

    def on_fit_end(self, params_dict):
        total_time = params_dict[ParamsDict.TOTAL_TIME]
        # final message
        self.logger.info(
            f"""\nEnd of training. Total time: {total_time:0.5f} seconds"""
        )

    def render_text(self, update_dict):
        text_format = ""

        for metric in update_dict:
            if metric != ParamsDict.HISTORY_LR:
                train_metric = update_dict[metric]["train"][-1]
                val_metric = update_dict[metric]["validation"][-1]

                if text_format:  # not empty
                    text_format = (
                        f"{text_format} • {metric} -> Train: "
                        f"{train_metric:.{self.prec}e} | "
                        f"Validation: {val_metric:.{self.prec}e}"
                    )
                else:
                    text_format = (
                        f"{metric} -> Train: "
                        f"{train_metric:.{self.prec}e} | Validation: "
                        f"{val_metric:.{self.prec}e}"
                    )
            else:
                if self.log_lr:
                    text_format = (
                        f"{text_format} • Learning Rate: "
                        f"{update_dict[metric][-1]}"
                    )

        return text_format


class StochasticWeightAveraging(Callback):
    """Applies a stochastic weight averaging to the training process.

    If you were to use a learning rate scheduler in addition to stochastic
    averaging, you must pass both to the constructor of this class instead of
    creating an individual callback for the standard lr scheduler. See
    `Examples` section.

    Parameters
    -----------
    swa_scheduler : torch.optim.lr_scheduler.Scheduler
        Stochastic weight averaging scheduler.
    start_epoch : int
        Epoch from which to start applying stochastic weight average.
    scheduler : torch.optim.lr_scheduler._LRScheduler
        Scheduler use to perform the lr reduction.
    metric : str, optional, default : torchfitter.conventions.ParamsDict.LOSS
        Metric to track in order to reduce the learning rate. If you want to
        use one of the passed metrics you must pass the class name. See
        examples section.
    on_train : bool, optional, default: True
        Whether to watch the train version of the metric (True) or the
        validation version of the metric (False)

    References
    ----------
    .. [1] Pavel Izmailov, Andrew Gordon Wilson and Vincent Queneneville-Belair
       PyTorch 1.6 now includes Stochastic Weight Averaging
       https://pytorch.org/blog/pytorch-1.6-now-includes-stochastic-weight-averaging/

    .. [2] PyTorch - Stochastic Weight Averaging
       https://pytorch.org/docs/stable/optim.html#stochastic-weight-averaging

    Examples
    --------
    >>> optimizer, model, criterion = ...
    >>> swa_model = torch.optim.swa_utils.AveragedModel(model)
    >>> swa_start = 160
    >>> scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
    ...    optimizer, T_max=300)
    >>> swa_scheduler = SWALR(optimizer, swa_lr=0.05)
    >>> swa_callback = StochasticWeightAveraging(
    ...     swa_scheduler, swa_start, scheduler=scheduler)
    >>> trainer = Trainer(callbacks=[swa_callback], **kwargs)
    >>> history = trainer.fit(...)

    Now we can get the SWA model by simply calling:
    >>> swa_model = swa_callback.get_swa_model()
    """

    def __init__(
        self,
        swa_scheduler: torch.optim.swa_utils.SWALR,
        start_epoch: int,
        scheduler: torch.optim.lr_scheduler._LRScheduler = None,
        metric: str = None,
        on_train: bool = True,
    ):
        super(StochasticWeightAveraging, self).__init__()

        self.swa_scheduler = swa_scheduler
        self.start_epoch = start_epoch
        self.scheduler = scheduler
        self.metric = metric
        self.on_train = on_train

        self.__swa_model = None

    def __repr__(self) -> str:
        return (
            f"StochasticWeightAveraging(swa_scheduler={self.swa_scheduler}, "
            f"start_epoch={self.start_epoch})"
        )

    def on_fit_start(self, params_dict: dict) -> None:
        model = params_dict[ParamsDict.MODEL]
        accelerator = params_dict[ParamsDict.ACCELERATOR]

        self.scheduler = accelerator.prepare(self.scheduler)
        self.swa_scheduler = accelerator.prepare(self.swa_scheduler)
        self.__swa_model = AveragedModel(model)

    def on_train_step_end(self, params_dict: dict) -> None:
        self.__run_scheduler(params_dict=params_dict)

    def __run_scheduler(self, params_dict: dict) -> None:
        """
        Run appropiate scheduler depending on the epoch number.
        """
        epoch = params_dict[ParamsDict.EPOCH_NUMBER]
        model = params_dict[ParamsDict.MODEL]

        if epoch >= self.start_epoch:
            # run swa scheduler
            self.__swa_model.update_parameters(model)
            self.swa_scheduler.step()
        else:
            # run "standard" scheduler
            if self.metric is not None:
                key = "train" if self.on_train else "validation"
                epoch_hist = params_dict[ParamsDict.EPOCH_HISTORY]
                metric = epoch_hist[self.metric][key][-1]
                self.scheduler.step(metric)
            else:
                self.scheduler.step()

    def on_fit_end(self, params_dict: dict) -> None:
        train_loader = params_dict[ParamsDict.TRAIN_LOADER]
        torch.optim.swa_utils.update_bn(train_loader, self.__swa_model)

    def get_swa_model(self) -> torch.nn.Module:
        """
        Getter method for the Stochastic Weight Averaged model.

        Returns
        -------
        swa_model : torch.nn.Module
            SWA model.
        """
        return self.__swa_model


class L1Regularization(Callback):
    """Applies L1 regularization over the model parameters.

    L1 is usually called 'Lasso Regression' (Least Absolute Shrinkage and
    Selection Operator). This callbacks is only applied to the train loss.

    Parameters
    ----------
    regularization_rate : float
        Regularization rate. Also called `lambda`.
    biases : bool, optional, default: False
        Whether to apply regularization over bias terms (True) or not (False).

    Note
    ----
    The penalty term already handles the product by the lambda regularization
    rate.

    """

    def __init__(self, regularization_rate: float, biases: bool = False):
        super().__init__()

        self.rate = regularization_rate
        self.biases = biases

    def on_loss_step_end(self, params_dict: dict) -> None:
        batch_tr_loss = params_dict[ParamsDict.BATCH_TRAIN_LOSS]
        device = params_dict[ParamsDict.DEVICE]
        model = params_dict[ParamsDict.MODEL]

        # Initialize with tensor, cannot be scalar
        penalty_term = torch.zeros(1, 1, requires_grad=True).to(device)

        for name, param in model.named_parameters():
            if not self.biases and name.endswith("bias"):
                continue

            penalty_term = penalty_term + param.norm(p=1)

        total_penalty = self.rate * penalty_term
        loss = total_penalty + batch_tr_loss

        # set loss
        params_dict[ParamsDict.BATCH_TRAIN_LOSS] = loss


class L2Regularization(Callback):
    """Applies L2 regularization over the model parameters.

    L2 is usually called 'Ridge Regression'. This callbacks is only applied to
    the train loss.

    Parameters
    ----------
    regularization_rate : float
        Regularization rate. Also called `lambda`.
    biases : bool, optional, default: False
        Whether to apply regularization over bias terms (True) or not (False).

    Note
    ----
    The penalty term already handles the product by the lambda regularization
    rate.

    """

    def __init__(self, regularization_rate: float, biases: bool = False):
        super().__init__()

        self.rate = regularization_rate
        self.biases = biases

    def on_loss_step_end(self, params_dict: dict) -> None:
        batch_tr_loss = params_dict[ParamsDict.BATCH_TRAIN_LOSS]
        device = params_dict[ParamsDict.DEVICE]
        model = params_dict[ParamsDict.MODEL]

        # Initialize with tensor, cannot be scalar
        penalty_term = torch.zeros(1, 1, requires_grad=True).to(device)

        for name, param in model.named_parameters():
            if not self.biases and name.endswith("bias"):
                continue

            penalty_term = penalty_term + param.norm(p=2)

        total_penalty = self.rate * penalty_term
        loss = total_penalty + batch_tr_loss

        # set loss
        params_dict[ParamsDict.BATCH_TRAIN_LOSS] = loss


class ElasticNetRegularization(Callback):
    r"""Linear combination of L1 and L2.

    According to [1], the lasso penalty is somewhat indifferent to the choice
    among a set of strong but correlated variables. The ridge penalty, on the
    other hand, tends to shrink the coefficients of correlated variables toward
    each other. Elastic net combines both using a weighting factor:

    .. math::

        \sum_{j=1}^{p} ( \alpha |\beta_{j}| + (1 + \alpha) \beta_{j}^{2} )

    Parameters
    ----------
    regularization_rate : float
        Regularization rate. Also called `lambda`.
    alpha : float
        Parameter to determine the mix of the penalties.
    biases : bool, optional, default: False
        Whether to apply regularization over bias terms (True) or not (False).

    Note
    ----
    The penalty term already handles the product by the lambda regularization
    rate.

    References
    ----------
    .. [1] Trevor Hastie, Robert Tibshirani, Jerome Friedman - The Elements of
       Statistical Learning.

    """

    def __init__(
        self, regularization_rate: float, alpha: float, biases: bool = False
    ):
        super().__init__()

        self.rate = regularization_rate
        self.biases = biases
        self.alpha = alpha

    def on_loss_step_end(self, params_dict: dict) -> None:
        batch_tr_loss = params_dict[ParamsDict.BATCH_TRAIN_LOSS]
        device = params_dict[ParamsDict.DEVICE]
        model = params_dict[ParamsDict.MODEL]

        # Initialize with tensor, cannot be scalar
        penalty_term = torch.zeros(1, 1, requires_grad=True).to(device)

        for name, param in model.named_parameters():
            if not self.biases and name.endswith("bias"):
                continue

            l1 = param.norm(p=1)
            l2 = param.norm(p=2)
            penalty_term = penalty_term + (
                self.alpha * l1 + (1 - self.alpha) * l2
            )

        total_penalty = self.rate * penalty_term
        loss = total_penalty + batch_tr_loss

        # set loss
        params_dict[ParamsDict.BATCH_TRAIN_LOSS] = loss



""" Base callbacks class """
import logging
from abc import ABC
from typing import List

from torchfitter.utils.convenience import get_logger

__all__ = ["Callback", "CallbackHandler"]


class Callback(ABC):
    """Base callbacks class.

    A callback allows to interact with the model along various relevant points
    during the training process. Each point is called hook, and each method of
    a callback allows to "attach" functionality to that particular hook.

    For example, if one were to run a method at the start of the fitting
    process he or she would pass a callback with the desired functionality
    filling the method "on_fit_start".

    Attributes
    ----------
    logger : logging.Logger
        Callback logger. You can set the logging level with the
        'set_log_level'.

    References
    ----------
    .. [1] Keras - keras.callbacks
       https://github.com/keras-team/keras/blob/master/keras/callbacks.py#L609
    """

    def __init__(self):
        self.log_name: str = "Callback"
        self.logger: logging.Logger = get_logger(name=self.log_name)
        level: int = self.logger.level
        logging.basicConfig(level=level)

    def set_log_level(self, log_level) -> None:
        """
        Set the logging level this callback instance.

        Parameters
        ----------
        log_level : int
            Logging level.
        """
        self.logger.setLevel(level=log_level)
        logging.basicConfig(level=log_level)

    def on_train_step_start(self, params_dict: dict) -> None:
        """Called at the start of a training step.

        A train step will involve the processing of all train batches included
        in the train dataloader.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        pass

    def on_train_step_end(self, params_dict: dict) -> None:
        """Called at the end of a training step.

        A train step will involve the processing of all train batches included
        in the train dataloader.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        pass

    def on_train_batch_start(self, params_dict: dict) -> None:
        """Called at the start of a batch train step.

        A batch train step will involve the processing of all samples in a
        single train batch.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.

        """
        pass

    def on_train_batch_end(self, params_dict: dict) -> None:
        """Called at the end of a batch train step.

        A batch train step will involve the processing of all samples in a
        single train batch.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.

        """
        pass

    def on_validation_step_start(self, params_dict: dict) -> None:
        """Called at the start of a validation step.

        A validation step will involve the processing of all validation batches
        included in the validation dataloader.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        pass

    def on_validation_step_end(self, params_dict: dict) -> None:
        """Called at the end of a validation step.

        A validation step will involve the processing of all validation batches
        included in the validation dataloader.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        pass

    def on_validation_batch_start(self, params_dict: dict) -> None:
        """Called at the start of a batch validation step.

        A batch validation step will involve the processing of all samples in a
        single validation batch.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.

        """
        pass

    def on_validation_batch_end(self, params_dict: dict) -> None:
        """Called at the end of a batch validation step.

        A batch validation step will involve the processing of all samples in a
        single validation batch.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.

        """
        pass

    def on_epoch_start(self, params_dict: dict) -> None:
        """Called at the start of an epoch.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        pass

    def on_epoch_end(self, params_dict: dict) -> None:
        """Called at the end of an epoch.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        pass

    def on_fit_start(self, params_dict: dict) -> None:
        """Called at the start of the fitting process.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        pass

    def on_fit_end(self, params_dict: dict) -> None:
        """Called at the end of the fitting process.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        pass

    def on_loss_step_begin(self, params_dict: dict) -> None:
        """Called at the start of the loss step.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        pass

    def on_loss_step_end(self, params_dict: dict) -> None:
        """Called at the end of the loss step.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        pass


class CallbackHandler(Callback):
    """Trainer callback handler.

    Class to handle callbacks during the training process. This class is itself
    a callback.

    Parameters
    ----------
    callbacks_list : list of torchfitter.callback.Callback
        List of callbacks to handle.
    """

    def __init__(self, callbacks_list):
        self.handle_callbacks: bool = True

        if callbacks_list is None:
            self.handle_callbacks = False
        elif not isinstance(callbacks_list, list):
            raise TypeError("Callbacks must be a list of callbacks")

        self.callbacks_list: List[Callback] = callbacks_list

    def set_log_level(self, log_level: int) -> None:
        """
        Set the logging level for all callbacks contained in this instance of
        CallbacksHandler.

        Parameters
        ----------
        log_level : int
            Logging level.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.set_log_level(log_level=log_level)

    def on_train_step_start(self, params_dict: dict) -> None:
        """Called at the start of a training step.

        Call this method for all given callbacks list. Any returned values will
        be ignored by the trainer.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_train_step_start(params_dict)

    def on_train_step_end(self, params_dict: dict) -> None:
        """Called at the end of a training step.

        Call this method for all given callbacks list. Any returned values will
        be ignored by the trainer.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_train_step_end(params_dict)

    def on_train_batch_start(self, params_dict: dict) -> None:
        """Called at the start of a batch train step.

        A batch train step will involve the processing of all samples in a
        single train batch.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.

        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_train_batch_start(params_dict)

    def on_train_batch_end(self, params_dict: dict) -> None:
        """Called at the end of a batch train step.

        A batch train step will involve the processing of all samples in a
        single train batch.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.

        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_train_batch_end(params_dict)

    def on_validation_step_start(self, params_dict: dict) -> None:
        """Called at the start of a validation step.

        Call this method for all given callbacks list. Any returned values will
        be ignored by the trainer.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_validation_step_start(params_dict)

    def on_validation_step_end(self, params_dict: dict) -> None:
        """Called at the end of a validation step.

        Call this method for all given callbacks list. Any returned values will
        be ignored by the trainer.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_validation_step_end(params_dict)

    def on_validation_batch_start(self, params_dict: dict) -> None:
        """Called at the start of a batch validation step.

        A batch validation step will involve the processing of all samples in a
        single validation batch.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.

        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_validation_batch_start(params_dict)

    def on_validation_batch_end(self, params_dict: dict) -> None:
        """Called at the end of a batch validation step.

        A batch validation step will involve the processing of all samples in a
        single validation batch.

        Subclasses should override for any actions to run. The trainer ignores
        any returned values from this function.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.

        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_validation_batch_end(params_dict)

    def on_epoch_start(self, params_dict: dict) -> None:
        """Called at the start of an epoch.

        Call this method for all given callbacks list. Any returned values will
        be ignored by the trainer.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_epoch_start(params_dict)

    def on_epoch_end(self, params_dict: dict) -> None:
        """Called at the end of an epoch.

        Call this method for all given callbacks list. Any returned values will
        be ignored by the trainer.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_epoch_end(params_dict)

    def on_fit_start(self, params_dict: dict) -> None:
        """Called at the start of the fitting process.

        Call this method for all given callbacks list. Any returned values will
        be ignored by the trainer.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_fit_start(params_dict)

    def on_fit_end(self, params_dict: dict) -> None:
        """Called at the end of the fitting process.

        Call this method for all given callbacks list. Any returned values will
        be ignored by the trainer.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_fit_end(params_dict)

    def on_loss_step_begin(self, params_dict: dict) -> None:
        """Called at the start of the loss step.

        Call this method for all given callbacks list. Any returned values will
        be ignored by the trainer.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_loss_step_begin(params_dict)

    def on_loss_step_end(self, params_dict: dict) -> None:
        """Called at the end of the loss step.

        Call this method for all given callbacks list. Any returned values will
        be ignored by the trainer.

        Parameters
        ----------
        params_dict : dict
            Dictionary containing the parameters of the training process.
        """
        if self.handle_callbacks:
            for callback in self.callbacks_list:
                callback.on_loss_step_end(params_dict)



from setuptools import setup, find_packages
from os import path
import versioneer


here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="torchfitter",
    version=versioneer.get_version(),
    description="Trainer to optimize PyTorch models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    author="Alejandro Pérez-Sanjuán",
    python_requires=">=3.7,",
    install_requires=[
        "rich",
        "numpy>=1.20.0",
        "accelerate>=0.11.0",
        "scikit-learn",
        "torchmetrics",
        "torch>=1.1.0",
    ],
)



""" Pool of miscellaneous and convenient functions. """

import logging

import torch

__all__ = [
    "check_model_on_cuda",
    "get_logger",
    "freeze_model",
    "unfreeze_model",
]


def check_model_on_cuda(model: torch.nn.Module) -> bool:
    """
    Check if the model is stored in a cuda device.

    Parameters
    ----------
    model : torch.nn.Module
        PyTorch model to check.

    Return
    ------
    bool
        True if the model is stored on a cuda device.
    """
    return next(model.parameters()).is_cuda


def get_logger(name: str, level: int = logging.INFO) -> logging.Logger:
    """
    Generate a logger with the specified name and level.

    Parameters
    ----------
    name : str
        Logger name.
    level : int
        Logging level for the logger.

    Returns
    -------
    logger : logging.Logger
        Logger.
    """
    logger = logging.getLogger(name=name)
    logger.setLevel(level=level)
    return logger


def freeze_model(model: torch.nn.Module) -> None:
    """Freeze the given model.

    This function is an inplace operations that deactivates the gradient of all
    parameters.

    Parameters
    ----------
    models : torch.nn.Module
        Model to freeze.
    """
    for param in model.parameters():
        param.requires_grad = False


def unfreeze_model(model: torch.nn.Module) -> None:
    """Unfreeze the given model.

    This function is an inplace operations that activates the gradient of all
    parameters.

    Parameters
    ----------
    models : torch.nn.Module
        Model to unfreeze.
    """
    for param in model.parameters():
        param.requires_grad = True



""" Utilities for the training process. """
from typing import Dict, List

import torch
import torchmetrics

from torchfitter.conventions import ParamsDict


class TrainerInternalState:
    """
    Class that keeps track of the trainer internal state during the fitting
    process.

    Parameters
    ----------
    model : torch.nn.Module
        The model to train.
    accelerator : accelerate.Accelerator
        Accelerator object.
    optimizer : torch.optim.Optimizer
        Algorithm used to optimize the model.

    Attributes
    ----------
    training_loss : float
        The current training loss.
    validation_loss : float
        The current validation loss.
    epoch_time : float
        The time it took to compute the current epoch.
    epoch_number : int
        The corresponding number of the current epoch.
    total_epochs : int
        The total number of epochs.
    total_time : float
        The total time it took to complete all epochs.
    stop_training : bool
        The total time it took to complete all epochs.
    history : dict
        Dictionary containing the metrics:
        * train_loss : list of floats
            Train loss for each epoch up to the current epoch.
        * validation_loss : list of floats
            Validation loss for each epoch up to the current epoch.
        * learning_rate : list of floats
            Learning rate for each epoch up to the current epoch.

        The dictionary can be updated with more keys through the method
        'add_history_element'.
    progress_bar : tqdm.tqdm
        Progress bar from tqdm library.
    """

    def __init__(self, model, accelerator, optimizer) -> None:
        self.___initialize_dict(
            model=model, accelerator=accelerator, optimizer=optimizer
        )

    def ___initialize_dict(self, model, accelerator, optimizer):
        """
        Helper function that initializes all attributes.
        """
        self.__dict__[ParamsDict.TRAIN_LOSS] = float("inf")
        self.__dict__[ParamsDict.VAL_LOSS] = float("inf")
        self.__dict__[ParamsDict.BATCH_TRAIN_LOSS] = float("inf")
        self.__dict__[ParamsDict.BATCH_VAL_LOSS] = float("inf")
        self.__dict__[ParamsDict.EPOCH_TIME] = 0
        self.__dict__[ParamsDict.EPOCH_NUMBER] = 1
        self.__dict__[ParamsDict.TOTAL_EPOCHS] = None
        self.__dict__[ParamsDict.TOTAL_TIME] = 0
        self.__dict__[ParamsDict.STOP_TRAINING] = False
        self.__dict__[ParamsDict.ACCELERATOR] = accelerator
        self.__dict__[ParamsDict.DEVICE] = accelerator.device
        self.__dict__[ParamsDict.MODEL] = model
        self.__dict__[ParamsDict.OPTIMIZER] = optimizer
        self.__dict__[ParamsDict.EPOCH_HISTORY] = {
            ParamsDict.LOSS: {"train": [], "validation": []},
            ParamsDict.HISTORY_LR: [],
        }
        self.__dict__[ParamsDict.BATCH_HISTORY] = {
            ParamsDict.LOSS: {"train": [], "validation": []},
            ParamsDict.HISTORY_LR: [],
        }

    def add_metrics(self, *args) -> None:
        """
        Add a metric to the history dictionary taking into account that metrics
        are computed both for training and validation steps.

        Parameters
        ----------
        *args : iterable
            Keys to add to the history dictionary.
        """
        for key in args:
            if not isinstance(key, str):
                raise TypeError(
                    f"'key' {key} must be a string, not {type(key)}"
                )

            for metric_type in [
                ParamsDict.EPOCH_HISTORY,
                ParamsDict.BATCH_HISTORY,
            ]:
                self.__dict__[metric_type][key] = {}  # avoid key error
                self.__dict__[metric_type][key]["train"] = []
                self.__dict__[metric_type][key]["validation"] = []

    def get_single_param(self, key: str) -> object:
        """
        Retrieve a single value from the set of internal attributes of the
        class.

        Parameters
        ----------
        key : str
            Parameter to retrieve.

        Returns
        -------
        object
            Value of passed 'key'.

        Raises
        ------
        KeyError
            If the requested attribute does not exist.
        """
        if key in self.__dict__:
            return self.__dict__[key]
        else:
            raise KeyError(f"Passed attribute '{key}' not found.")

    def get_state_dict(self) -> dict:
        """
        Return the internal dictionary that stores all parameters.

        Returns
        -------
        dict
            Internal dictionary.
        """
        return self.__dict__

    def get_params(self, *args) -> List[object]:
        """
        Parameters
        ----------
        *args : tuple
            Iterable of strings containing the keys to retrieve.

        Returns
        -------
        params : list
            List of values for each passed argument. The order is kept.

        Raises
        ------
        TypeError
            If the passed arguments are not strings.
        """
        params = []
        for key in args:
            if not isinstance(key, str):
                raise TypeError("Key to retrieve must be a string")

            val = self.get_single_param(key=key)
            params.append(val)

        return params

    def update_loss_history(
        self, value: float, is_train: bool = True, is_batch: bool = False
    ) -> None:
        """
        Update history loss dictionary with the passed value.

        Parameters
        ----------
        value : float
            Value to update the learning rate.
        is_train : bool, optional, default: True
            Whether to update the train history (True) or the validation
            history (False).
        is_batch : bool, optional, default: False
            Whether to update the batch history (True) or the epoch history
            (False).
        """
        # select batch or epoch
        hist_type = (
            ParamsDict.BATCH_HISTORY if is_batch else ParamsDict.EPOCH_HISTORY
        )

        # select train or validation
        tr_or_val = "train" if is_train else "validation"

        # update
        self.__dict__[hist_type][ParamsDict.LOSS][tr_or_val].append(value)

    def update_lr_history(self, value: float, is_batch: bool = False) -> None:
        """
        Update learning rate history with the passed value.

        Parameters
        ----------
        value : float
            Value to update the learning rate.
        is_batch : bool, optional, default: False
            Whether to update the batch history (True) or the epoch history
            (False).
        """
        # select batch or epoch
        hist_type = (
            ParamsDict.BATCH_HISTORY if is_batch else ParamsDict.EPOCH_HISTORY
        )

        # update lr
        self.__dict__[hist_type][ParamsDict.HISTORY_LR].append(value)

    def update_params(self, **kwargs) -> None:
        """
        Update internal state with the passed key-value pairs.

        Parameters
        ----------
        kwargs : dict
            Dictionary with keys to update.
        """
        for key, value in kwargs.items():
            self.__dict__[key] = value

    def update_metrics(
        self, is_train: bool, is_batch: bool = False, **kwargs
    ) -> None:
        """
        Update the metrics dictionary with the passed key-value pairs.

        Parameters
        ----------
        is_train : bool
            If True, the train metrics will be updated. Otherwise, the
            validation metrics will be updated.
        """
        # select batch or epoch
        hist_type = (
            ParamsDict.BATCH_HISTORY if is_batch else ParamsDict.EPOCH_HISTORY
        )

        # select train or validation
        tr_or_val = "train" if is_train else "validation"

        for metric, value in kwargs.items():
            if metric not in self.__dict__[hist_type]:
                raise KeyError(f"'{metric}' not found in history dict.")

            self.__dict__[hist_type][metric][tr_or_val].append(value)

    def reset_parameters(self, reset_model=True):
        model = self.__dict__[ParamsDict.MODEL]
        optim = self.__dict__[ParamsDict.OPTIMIZER]
        accelerator = self.__dict__[ParamsDict.ACCELERATOR]

        if reset_model:
            model.reset_parameters()

        self.___initialize_dict(
            model=model, accelerator=accelerator, optimizer=optim
        )


class MetricsHandler:
    """Class to handle the metrics computation in the process.

    The metrics must be provided using the torchmetrics package API; that is,
    the metrics modules must support single and accumulated batch computation.
    See references for more information.

    Parameters
    ----------
    metrics_list : list of torchmetrics.Metric
        A list of all the metrics to be computed.
    criterion : torch.nn.Module
        Loss function criterion used to optimize the model.

    Attributes
    ----------
    metric_names : list of str
        Names of the metrics. The names are automatically computed using
        type(metric).__name__

    References
    ----------
    .. [1] PyTorch-Lightning - torchmetrics
       https://torchmetrics.readthedocs.io/en/latest/

    """

    def __init__(
        self,
        metrics_list: List[torchmetrics.Metric],
        criterion: torch.nn.Module,
        device,
    ) -> None:

        self.metrics_list = metrics_list
        self.criterion = criterion
        self.device = device

        # handle metrics if there are metrics
        self.__handle_metrics = False if self.metrics_list is None else True

        if self.__handle_metrics:
            self.metric_names = [
                type(metric).__name__ for metric in self.metrics_list
            ]

            # move metrics to device
            metrics = [metric.to(self.device) for metric in self.metrics_list]
            self.metrics_list = metrics
        else:
            self.metric_names = None

    def single_batch_computation(
        self, predictions: torch.Tensor, target: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        """
        Perform a single batch computation on all metrics.

        Parameters
        ----------
        predictions : torch.Tensor
            Predictions of the model.
        target : torch.Tensor
            Desired output of the model.

        Returns
        -------
        results : list of torch.Tensor
            Results of each metric.
        """
        if self.__handle_metrics:
            results = {}
            for metric in self.metrics_list:
                key = type(metric).__name__
                value = metric(predictions, target).item()

                results[key] = value

            return results

    def accumulated_batch_computation(self) -> Dict[str, torch.Tensor]:
        """
        Perform the metric computation on all batches using custom
        accumulation.

        Returns
        -------
        results : list of torch.Tensor
            Results of each metric.
        """
        if self.__handle_metrics:
            results = {}
            for metric in self.metrics_list:
                key = type(metric).__name__
                value = metric.compute().item()

                results[key] = value

            return results

    def reset_metrics(self) -> None:
        """
        Reset the internal state of the metrics in order to prepare them to
        receive new data.
        """
        if self.__handle_metrics:
            for metric in self.metrics_list:
                metric.reset()



"""
Preprocessing functions.
"""
import math
from typing import Iterable, List, Union

import numpy as np
import torch
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import train_test_split as __tr_test_split

__all__ = [
    "numpy_to_torch",
    "train_test_val_split",
    "torch_to_numpy",
    "tabular_to_sliding_dataset",
]


def torch_to_numpy(tensor: torch.Tensor) -> np.ndarray:
    """
    Cast a torch.Tensor to a numpy.ndarray dealing with device management if
    any. For example, a tensor may need to be detached but it is not stored on
    the cpu.

    Parameters
    ----------
    tensor : torch.Tensor
        Tensor to convert to numpy.

    Returns
    -------
    array : numpy.array
        NumPy array.
    """
    try:
        array = tensor.detach().numpy()
    except Exception:
        array = tensor.cpu().detach().numpy()

    return array


def numpy_to_torch(array: np.ndarray, dtype: str) -> torch.Tensor:
    """
    Cast a numpy array to a torch tensor of the given dtype.

    Parameters
    ----------
    array : numpy.array
        Array to transform.
    dtype : str
        Desired data type.

    Returns
    -------
    tensor : torch.Tensor
        Torch tensor with the desired properties.

    Examples
    --------
    >>> import torch
    >>> import numpy as np
    >>> from torchfitter.utils.preprocessing import numpy_to_torch
    >>> arr = np.array([1,2,3], dtype='int')
    >>> arr.dtype
    dtype('int64')

    >>> tensor = numpy_to_torch(arr, dtype='long')
    >>> tensor.dtype
    torch.int64
    """
    return getattr(torch.from_numpy(array), dtype)()


def train_test_val_split(
    X: np.ndarray,
    y: np.ndarray,
    train_ratio: float = 0.70,
    validation_ratio: float = 0.20,
    test_ratio: float = 0.10,
    random_state: int = 42,
    shuffle: bool = False,
    stratify=None,
) -> Iterable[np.ndarray]:
    """
    Splits the given dataset into train, validation and test sets.

    This function applies `sklearn.model_selection.train_test_split` twice to
    generate the outputs.

    Parameters
    ----------
    X : array-like
        Features input set.
    y : array-like
        Labels input set.
    train_ratio : float, optional, default: 0.70
        Ratio of train set.
    validation_ratio : float, optional, default: 0.20
        Ratio of validation set.
    test_ratio : float, optional, default: 0.10
        Ratio of test set.
    random_state : int, optional, default: 42
        Controls the shuffling applied to the data before applying the split.
        Pass an int for reproducible output across multiple function calls.
    shuffle : bool, optional, default: False
        Whether or not to shuffle the data before splitting. If shuffle=False
        then stratify must be None. Shuffle will only be applied in the first
        split.
    stratify : array-like, default: None
        If not None, data is split in a stratified fashion, using this as the
        class labels. Stratify will only be applied in the first split.

    Returns
    -------
    X_train: numpy.ndarray
        Features train set.
    y_train: numpy.ndarray
        Labels train set.
    X_val: numpy.ndarray
        Features validation set.
    y_val: numpy.ndarray
        Labels validation set.
    X_test: numpy.ndarray
        Features test set.
    y_test : numpy.ndarray
        Labels test set.

    References
    ----------
    See `sklearn.model_selection.train_test_split`
    """
    msg = "'train_ratio', 'validation_ratio' and 'test_ratio' must sum 1"
    _sum = train_ratio + validation_ratio + test_ratio
    assert math.isclose(_sum, 1), msg

    test_size = 1 - train_ratio
    X_train, X_test, y_train, y_test = __tr_test_split(
        X,
        y,
        test_size=test_size,
        random_state=random_state,
        shuffle=shuffle,
        stratify=stratify,
    )

    val_size = test_ratio / (test_ratio + validation_ratio)
    X_val, X_test, y_val, y_test = __tr_test_split(
        X_test,
        y_test,
        test_size=val_size,
        random_state=random_state,
        shuffle=False,
        stratify=None,
    )

    return X_train, y_train, X_val, y_val, X_test, y_test


def tabular_to_sliding_dataset(
    dataset: np.ndarray,
    validation_idx: int,
    test_idx: int,
    n_past: int,
    n_future: int,
    make_writable: bool = True,
    scaler: Union[TransformerMixin, BaseEstimator] = None,
) -> List[np.ndarray]:
    """Convert a tabular or 2D dataset to a sliding window dataset (3D).

    This function expects a datatype that supports the array protocol.
    E.g.: Pandas DataFrame or NumPy arrays.

    Parameters
    ----------
    dataset : array-like
        Array-like object.
    validation_idx : int
        Index to create the validation set.
    test_idx : int
        Index to create the testing set.
    n_past : int
        Number of past steps to make predictions. It will be used to
        generate the features.
    n_future : int
        Number of future steps to predict. It will be used to generate
        the labels.
    make_writable : bool, optional, default: True
        Make the resulting arrays writable by creating a copy of the view.
    scaler : sklearn.base.TransformerMixin, optional, default: None
        If not None, the data will be normalized with the passed scaler.
        Assumes distribution does not vary over time.

    Returns
    -------
    output : list of numpy.ndarray
        A list containing the resulting arrays. They appear in this order:
            * X_train: Features train set.
            * y_train: Labels train set.
            * X_val: Features validation set.
            * y_val: Labels validation set.
            * X_test: Features test set.
            * y_test : Labels test set.

    Warning
    -------
    This function is very memory-consuming.

    See Also
    --------
    torchfitter.utils.preprocessing.train_test_val_split

    TODO
    ----
    * Allow spliting by percentage.
    * Allow single-feature forecasting instead of multi-forecasting.
    * Use `train_test_val_split` to abstract the splitting.
    * Allow selecting the target column.
    """

    def get_train_and_test(array, n_past, n_future):
        """
        Convenient sub-function that wraps to functionality to
        create a rolling view and select the past as features
        and the future as labels.
        """
        window_length = n_past + n_future
        roll_view = np.lib.stride_tricks.sliding_window_view(
            array, window_length, axis=0
        )
        X = roll_view[:, :, :n_past]
        y = roll_view[:, :, n_past:]
        return X, y

    # type-agnostic
    arr = dataset.__array__()
    if arr.ndim == 1:
        arr = arr.reshape(-1, 1)

    # split
    train = arr[:validation_idx]
    validation = arr[validation_idx:test_idx]
    test = arr[test_idx:]

    if scaler is not None:
        scaler.fit(train)

        train = scaler.transform(train)
        validation = scaler.transform(validation)
        test = scaler.transform(test)

    # get a rolling view of each data chunk
    output = []
    for chunk in [train, validation, test]:
        X, y = get_train_and_test(
            array=chunk, n_past=n_past, n_future=n_future
        )

        # make a copy to generate a writable array
        if make_writable:
            _tup = (X.copy(), y.copy())
        else:
            _tup = (X, y)

        output.append(_tup)

    # unpack and return
    output = [item for sublist in output for item in sublist]
    return output
